/*
 * File: getnumorrange.c
 *   By: David Hiebeler
 *       http://www.math.umaine.edu/~hiebeler
 *       November 2010
 *       Comma-separated lists added, August 2012
 *       Current version: 2013 Aug 19
 *
 * This file contains routines to read strings of the form:
 *   val
 *   begin:end:numVals
 *   begin:end::stepsize
 *   val1,val2,val3,val4,...  (comma-separated list of explicit values)
 *
Copyright (c) 2013 David Hiebeler. All Rights Reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

3. The name of the author may not be used to endorse or promote
products derived from this software without specific prior written
permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */

#include <stdio.h>
#include <string.h>

/*
 * Parse a character string, and read it as either a numerical value,
 * or a range of values, of the form
 *  begin:end:numVals
 * or
 *  begin:end::stepsize
 * or as a comma-separated list of values, i.e.
 *  val1,val2,val3,val4,...
 * Return 0 on success, non-zero on failure.
 */
int
getIntNumOrRange(char *s, int *gotValue, int *begin, int *end, int *numVals, 
		 int *stepSize, char **linkedParmPtr, int *iValList, int *gotValListPtr)
{
  int i=0, ii, j, foundColon=0, foundDblColon=0, foundComma=0, tmpNumVals, tmpVal, savedAtPos=-1;
  char *linkedParm = NULL, savedChar;

  for (i=0; i < strlen(s); i++) {
    if (s[i] == '@') {
      linkedParm = &s[i+1];
      savedAtPos = i;
      s[i] = '\0';
    }
  }
  *linkedParmPtr = linkedParm;
  for (i=0; i < strlen(s); i++) {
    if (s[i] == ':') {
      foundColon = 1;
      if (s[i+1] == ':')
	foundDblColon = 1;
    }
  }
  tmpNumVals = 1;
  for (i=0; i < strlen(s); i++) {
    if (s[i] == ',') {
      foundComma = 1;
      tmpNumVals++;
    }
  }
  if (foundComma && foundColon) {
    if (savedAtPos != -1)
      s[savedAtPos] = '@';
    return(1);
  }
  *gotValListPtr = 0;
  if ((foundColon == 0) && (foundComma == 0)) {
    *gotValue = 1;
    if (sscanf(s, "%d", begin) != 1) {
      if (savedAtPos != -1)
	s[savedAtPos] = '@';
      return(1);
    }
    else {
      *end = *begin;
      *numVals = 1;
      *stepSize = 0;
      if (savedAtPos != -1)
	s[savedAtPos] = '@';
      return(0);
    }
  } else if (foundComma) {
    *gotValListPtr = 1;
    *gotValue = 0;
    i = 0;
    ii = 0;
    for (j=0; j < tmpNumVals; j++) {
      /* read the next value from the list */
      while ((s[i] != ',') && (s[i] != '\0'))
	i++;
      savedChar = s[i];
      s[i] = '\0';
      if (sscanf(&s[ii], "%d", &iValList[j]) != 1) {
	if (savedAtPos != -1)
	  s[savedAtPos] = '@';
	return(1);
      }
      s[i] = savedChar;
      i++;
      ii = i;
    }
    *numVals = tmpNumVals;
    if (savedAtPos != -1)
      s[savedAtPos] = '@';
    return(0);
  } else {  /* found a colon */
    *gotValue = 0;
    i = 0;
    if (sscanf(s, "%d", begin) != 1) {
      if (savedAtPos != -1)
	s[savedAtPos] = '@';
      return(2);
    }
    while ((s[i] != '\0') && (s[i] != ':'))
      i++;
    if (s[i] == ':')
      i++;
    if (sscanf(&s[i], "%d", end) != 1) {
      if (savedAtPos != -1)
	s[savedAtPos] = '@';
      return(3);
    }
    while ((s[i] != '\0') && (s[i] != ':'))
      i++;
    if (s[i] == ':')
      i++;
    if (foundDblColon) {
      if (s[i] != ':') {
	if (savedAtPos != -1)
	  s[savedAtPos] = '@';
	return(4);
      }
      else
	i++;
      if (sscanf(&s[i], "%d", stepSize) != 1) {
	if (savedAtPos != -1)
	  s[savedAtPos] = '@';
	return(5);
      }
      tmpNumVals = 1;
      tmpVal = *begin + *stepSize;
      while (tmpVal <= *end) {
	tmpNumVals++;
	tmpVal += *stepSize;
      }
      *numVals = tmpNumVals; {
	if (savedAtPos != -1)
	  s[savedAtPos] = '@';
      return(0);
      }
    } else {  /* no double colon */
      if (sscanf(&s[i], "%d", numVals) != 1) {
	if (savedAtPos != -1)
	  s[savedAtPos] = '@';
	return(6);
      }
      if (*numVals < 2) {
	if (savedAtPos != -1)
	  s[savedAtPos] = '@';
	return(7);
      }
      if ((*end-*begin)%((*numVals) - 1)) {  /* numVals gives non-integer stepSize */
	if (savedAtPos != -1)
	  s[savedAtPos] = '@';
	return(8);
      }
      *stepSize = (*end-*begin)/((*numVals)-1);
      if (savedAtPos != -1)
	s[savedAtPos] = '@';
      return(0);
    }
  }
}


/*
 * Parse a character string, and read it as either a numerical value,
 * a range of values, or a comma-separated list of values.
 * Return 0 on success, non-zero on failure.
 */
int
getDoubleNumOrRange(char *s, int *gotValue, double *begin, double *end, 
		    int *numVals, double *stepSize, char **linkedParmPtr, double *dValList, int *gotValListPtr)
{
  int i=0, ii, j, foundColon=0, foundDblColon=0, foundComma=0, tmpNumVals, savedAtPos=-1;
  double tmpVal;
  char *linkedParm = NULL, savedChar;

  for (i=0; i < strlen(s); i++) {
    if (s[i] == '@') {
      linkedParm = &s[i+1];
      savedAtPos = i;
      s[i] = '\0';
    }
  }
  *linkedParmPtr = linkedParm;
  for (i=0; i < strlen(s); i++) {
    if (s[i] == ':') {
      foundColon = 1;
      if (s[i+1] == ':')
	foundDblColon = 1;
    }
  }
  tmpNumVals = 1;
  for (i=0; i < strlen(s); i++) {
    if (s[i] == ',') {
      foundComma = 1;
      tmpNumVals++;
    }
  }
  if (foundComma && foundColon) {
    if (savedAtPos != -1)
      s[savedAtPos] = '@';
    return(1);
  }
  *gotValListPtr = 0;
  if ((foundColon == 0) && (foundComma == 0)) {
    *gotValue = 1;
    if (sscanf(s, "%lg", begin) != 1) {
      if (savedAtPos != -1)
	s[savedAtPos] = '@';
      return(1);
    }
    else {
      *end = *begin;
      *numVals = 1;
      *stepSize = 0.0; 
      if (savedAtPos != -1)
	s[savedAtPos] = '@';
      return(0);
    }
  } else if (foundComma) {
    *gotValListPtr = 1;
    *gotValue = 0;
    i = 0;
    ii = 0;
    for (j=0; j < tmpNumVals; j++) {
      /* read the next value from the list */
      while ((s[i] != ',') && (s[i] != '\0'))
	i++;
      savedChar = s[i];
      s[i] = '\0';
      if (sscanf(&s[ii], "%lg", &dValList[j]) != 1) {
	if (savedAtPos != -1)
	  s[savedAtPos] = '@';
	return(1);
      }
      s[i] = savedChar;
      i++;
      ii = i;
    }
    *numVals = tmpNumVals;
    if (savedAtPos != -1)
      s[savedAtPos] = '@';
    return(0);
  } else {  /* found a colon */
    *gotValue = 0;
    i = 0;
    if (sscanf(s, "%lg", begin) != 1) {
      if (savedAtPos != -1)
	s[savedAtPos] = '@';
      return(2);
    }
    while ((s[i] != '\0') && (s[i] != ':'))
      i++;
    if (s[i] == ':')
      i++;
    if (sscanf(&s[i], "%lg", end) != 1) {
      if (savedAtPos != -1)
	s[savedAtPos] = '@';
      return(3);
    }
    while ((s[i] != '\0') && (s[i] != ':'))
      i++;
    if (s[i] == ':')
      i++;
    if (foundDblColon) {
      if (s[i] != ':') {
	if (savedAtPos != -1)
	  s[savedAtPos] = '@';
	return(4);
      }
      else
	i++;
      if (sscanf(&s[i], "%lg", stepSize) != 1) {
	if (savedAtPos != -1)
	  s[savedAtPos] = '@';
	return(5);
      }
      tmpNumVals = 1;
      tmpVal = *begin + *stepSize;
      while (tmpVal <= *end) {
	tmpNumVals++;
	tmpVal = *begin + tmpNumVals*(*stepSize);
      }
      *numVals = tmpNumVals;
      if (savedAtPos != -1)
	s[savedAtPos] = '@';
      return(0);
    } else {  /* no double colon */
      if (sscanf(&s[i], "%d", numVals) != 1) {
	if (savedAtPos != -1)
	  s[savedAtPos] = '@';
	return(6);
      }
      if (*numVals < 2) {
	if (savedAtPos != -1)
	  s[savedAtPos] = '@';
	return(7);
      }
      *stepSize = (*end-*begin)/((*numVals)-1.0);
      if (savedAtPos != -1)
	s[savedAtPos] = '@';
      return(0);
    }
  }
}


