/*
 * File: parmloop.c
 * By: David Hiebeler
 *     http://www.math.umaine.edu/~hiebeler
 *     First written: May 2011
 *     Current version: 2013 Aug 19
 *
 * Facilities for looping over parameter values
 * 
 * Parameter values can take the form:
 *   begin:end:numVals
 *   begin:end::stepSize
 *   val1,val2,val3,val4...   (i.e. a comma-separated list of explicit values)
 * Any of the above can have an "@parm" suffix to link the parameter with the specified
 * prior parameter, e.g. -phi 2:4:3@alpha to have phi go from 2 to 4 using 3 values,
 *     but have the parameter linked to the other parameter alpha, or -mu .77,.32,.19@alpha
 *
Copyright (c) 2013 David Hiebeler. All Rights Reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

3. The name of the author may not be used to endorse or promote
products derived from this software without specific prior written
permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "getnumorrange.h"
#include "parmlooptypes.h"

#define INT 1
#define DOUBLE 2

/*
 * Initialize data structure
 */
void
PLinitData(PLdata *pldptr)
{
  pldptr->numParmsToLoopOver=0;
  pldptr->numLinkedParms=0;
  pldptr->totalParmCombos=1;
}


/*
 * Read a string specifying a parameter range of type double.
 * str: string to read from
 * pldptr: pointer to PLdata holding parameter info
 * varptr: pointer to the actual variable to be used
 * name: the name you refer to when making another parameter that links to this one;
 *       it is also used when displaying/printing the parameter.
 */
int
PLreadDoubleArg(char *str, PLdata *pldptr, double *varptr, char *name)
{
  int retVal, gotValue, numVals, i, gotList;
  double begin, end, stepSize, dValList[MAXVALLISTLEN];
  char *linkedParmName;
  PLdata *newpldptr;

  retVal = getDoubleNumOrRange(str, &gotValue, &begin, &end, &numVals, &stepSize, &linkedParmName, dValList, &gotList);
  if (retVal == 7) {
    fprintf(stderr, "a:b:n parameter indicated only one step, but you must use a minimum of two!\n");
    return(1);
  }
  else if (retVal > 0)
    return(1);
  else if (gotValue) {  /* just got a single value, not a range */
    *varptr = begin;
    return(0);
  } else {
    if (linkedParmName != NULL) {
      for (i=0; i < pldptr->numParmsToLoopOver; i++)
	if (!strcmp(pldptr->parmName[i], linkedParmName)) {
	  if ((newpldptr=(PLdata *)malloc(sizeof(PLdata)))==NULL) {
	    fprintf(stderr, "PLreadDoubleArg(): Unable to malloc newpldptr");
	    exit(1);
	  }
	  PLinitData(newpldptr);
	  if (pldptr->numVals[i] != numVals) {
	    fprintf(stderr, "Numbers of arguments in linked parameters `%s' and `%s' did not agree!\n", name, linkedParmName);
	    exit(1);
	  }
	  newpldptr->linkedParm[0] = pldptr->linkedParm[i];
	  pldptr->linkedParm[i] = newpldptr;
	  pldptr->numLinkedParms++;
	  pldptr = newpldptr;
	  break;
	}
    }
    else
      pldptr->linkedParm[pldptr->numParmsToLoopOver] = NULL;
    pldptr->numVals[pldptr->numParmsToLoopOver] = numVals;
    pldptr->parmType[pldptr->numParmsToLoopOver] = DOUBLE;
    pldptr->parmPtr[pldptr->numParmsToLoopOver] = varptr;
    pldptr->parmName[pldptr->numParmsToLoopOver] = name;
    if (gotList) {
      for (i=0; i < numVals; i++)
	pldptr->dValues[pldptr->numParmsToLoopOver][i] = dValList[i];
      pldptr->haveValueList[pldptr->numParmsToLoopOver] = 1;
    } else {  /* didn't get explicit list of values */
      pldptr->dbegin[pldptr->numParmsToLoopOver] = begin;
      pldptr->dend[pldptr->numParmsToLoopOver] = end;
      pldptr->dstepSize[pldptr->numParmsToLoopOver] = stepSize;
      pldptr->haveValueList[pldptr->numParmsToLoopOver] = 0;
    }
    if (linkedParmName == NULL) {
      pldptr->totalParmCombos *= pldptr->numVals[pldptr->numParmsToLoopOver];
      pldptr->numParmsToLoopOver++;
    }
    return(0);
  }
}


/*
 * Read a string specifying a parameter range of type int
 * str: string to read from
 * pldptr: pointer to PLdata holding parameter info
 * varptr: pointer to the actual variable to be used
 * name: the name you refer to when making another parameter that links to this one;
 *       it is also used when displaying/printing the parameter.
 */
int
PLreadIntArg(char *str, PLdata *pldptr, int *varptr, char *name)
{
  int retVal, gotValue, numVals, i, gotList;
  int begin, end, stepSize, iValList[MAXVALLISTLEN];
  char *linkedParmName;
  PLdata *newpldptr;

  retVal = getIntNumOrRange(str, &gotValue, &begin, &end, &numVals, &stepSize, &linkedParmName, iValList, &gotList);
  if (retVal == 7) {
    fprintf(stderr, "a:b:n parameter indicated only one step, but you must use a minimum of two!\n");
    return(1);
  }
  if (retVal == 8) {
    fprintf(stderr, "number of values for swept parameter gives non-integer stepSize!\n");
    return(1);
  }
  else if (retVal > 0)
    return(1);
  else if (gotValue) {  /* just got a single value, not a range */
    *varptr = begin;
    return(0);
  } else {
    if (linkedParmName != NULL) {
      for (i=0; i < pldptr->numParmsToLoopOver; i++)
	if (!strcmp(pldptr->parmName[i], linkedParmName)) {
	  if ((newpldptr=(PLdata *)malloc(sizeof(PLdata)))==NULL) {
	    fprintf(stderr, "PLreadDoubleArg(): Unable to malloc newpldptr");
	    exit(1);
	  }
	  PLinitData(newpldptr);
	  if (pldptr->numVals[i] != numVals) {
	    fprintf(stderr, "Numbers of arguments in linked parameters `%s' and `%s' did not agree!\n", name, linkedParmName);
	    exit(1);
	  }
	  newpldptr->linkedParm[0] = pldptr->linkedParm[i];
	  pldptr->linkedParm[i] = newpldptr;
	  pldptr->numLinkedParms++;
	  pldptr = newpldptr;
	  break;
	}
    }
    else
      pldptr->linkedParm[pldptr->numParmsToLoopOver] = NULL;
    pldptr->numVals[pldptr->numParmsToLoopOver] = numVals;
    pldptr->parmType[pldptr->numParmsToLoopOver] = INT;
    pldptr->parmPtr[pldptr->numParmsToLoopOver] = varptr;
    pldptr->parmName[pldptr->numParmsToLoopOver] = name;
    if (gotList) {
      for (i=0; i < numVals; i++)
	pldptr->iValues[pldptr->numParmsToLoopOver][i] = iValList[i];
      pldptr->haveValueList[pldptr->numParmsToLoopOver] = 1;
    } else {  /* didn't get explicit list of values */
      pldptr->ibegin[pldptr->numParmsToLoopOver] = begin;
      pldptr->iend[pldptr->numParmsToLoopOver] = end;
      pldptr->istepSize[pldptr->numParmsToLoopOver] = stepSize;
      pldptr->haveValueList[pldptr->numParmsToLoopOver] = 0;
    }
    if (linkedParmName == NULL) {
      pldptr->totalParmCombos *= pldptr->numVals[pldptr->numParmsToLoopOver];
      pldptr->numParmsToLoopOver++;
    }
    return(0);
  }
}


/*
 * Read a string specifying a parameter range of type int
 * This is for an argument that's a counter, e.g. "-reps 10" gets interpreted as "-reps 1:10::1"
 * str: string to read from
 * pldptr: pointer to PLdata holding parameter info
 * varptr: pointer to the actual variable to be used
 * name: the name you refer to when making another parameter that links to this one;
 *       it is also used when displaying/printing the parameter.
 */
int
PLreadIntArgValueAsCounter(char *str, PLdata *pldptr, int *varptr, char *name)
{
  int retVal, gotValue, numVals, i, gotList;
  int begin, end, stepSize, iValList[MAXVALLISTLEN];
  char *linkedParmName;
  PLdata *newpldptr;

  retVal = getIntNumOrRange(str, &gotValue, &begin, &end, &numVals, &stepSize, &linkedParmName, iValList, &gotList);
  if (retVal == 7) {
    fprintf(stderr, "a:b:n parameter indicated only one step, but you must use a minimum of two!\n");
    return(1);
  }

  if (retVal == 8) {
    fprintf(stderr, "number of values for swept parameter gives non-integer stepSize!\n");
    return(1);
  }
  else if (retVal > 0)
    return(1);
  else if (gotValue) {  /* just got a single value, not a range */
    begin = 1;
    numVals = end;
    stepSize = 1;
  }
  if (linkedParmName != NULL) {
    for (i=0; i < pldptr->numParmsToLoopOver; i++)
      if (!strcmp(pldptr->parmName[i], linkedParmName)) {
	if ((newpldptr=(PLdata *)malloc(sizeof(PLdata)))==NULL) {
	  fprintf(stderr, "PLreadDoubleArg(): Unable to malloc newpldptr");
	  exit(1);
	}
	PLinitData(newpldptr);
	if (pldptr->numVals[i] != numVals) {
	  fprintf(stderr, "Numbers of arguments in linked parameters `%s' and `%s' did not agree!\n", name, linkedParmName);
	  exit(1);
	}
	newpldptr->linkedParm[0] = pldptr->linkedParm[i];
	pldptr->linkedParm[i] = newpldptr;
	pldptr->numLinkedParms++;
	pldptr = newpldptr;
	break;
      }
  }
  else
    pldptr->linkedParm[pldptr->numParmsToLoopOver] = NULL;
  pldptr->numVals[pldptr->numParmsToLoopOver] = numVals;
  pldptr->parmType[pldptr->numParmsToLoopOver] = INT;
  pldptr->parmPtr[pldptr->numParmsToLoopOver] = varptr;
  pldptr->parmName[pldptr->numParmsToLoopOver] = name;
  if (gotList) {
    for (i=0; i < numVals; i++)
      pldptr->iValues[pldptr->numParmsToLoopOver][i] = iValList[i];
    pldptr->haveValueList[pldptr->numParmsToLoopOver] = 1;
  } else {  /* didn't get explicit list of values */
    pldptr->ibegin[pldptr->numParmsToLoopOver] = begin;
    pldptr->iend[pldptr->numParmsToLoopOver] = end;
    pldptr->istepSize[pldptr->numParmsToLoopOver] = stepSize;
    pldptr->haveValueList[pldptr->numParmsToLoopOver] = 0;
  }
  if (linkedParmName == NULL) {
    pldptr->totalParmCombos *= pldptr->numVals[pldptr->numParmsToLoopOver];
    pldptr->numParmsToLoopOver++;
  }
  return(0);
}


/*
 * Set a parameter variable to its value indicated in the PLdata struct
 */
void
PLsetParmValue(PLdata *pldptr, int parmInd, int parmCounterVal)
{
  if (pldptr->haveValueList[parmInd] == 0) {
    if (pldptr->parmType[parmInd] == DOUBLE)
      *((double *)pldptr->parmPtr[parmInd]) = pldptr->dbegin[parmInd] + pldptr->dstepSize[parmInd]*parmCounterVal;
    else if (pldptr->parmType[parmInd] == INT)
      *((int *)pldptr->parmPtr[parmInd]) = pldptr->ibegin[parmInd] + pldptr->istepSize[parmInd]*parmCounterVal;
  } else {
    if (pldptr->parmType[parmInd] == DOUBLE)
      *((double *)pldptr->parmPtr[parmInd]) = pldptr->dValues[parmInd][parmCounterVal];
    else if (pldptr->parmType[parmInd] == INT)
      *((int *)pldptr->parmPtr[parmInd]) = pldptr->iValues[parmInd][parmCounterVal];
  }
  if (pldptr->linkedParm[parmInd] != NULL)
    PLsetParmValue(pldptr->linkedParm[parmInd], 0, parmCounterVal);
}


/*
 * Initialize/set all parameter variables
 */
void
PLinitVars(PLdata *pldptr)
{
  int i;

  for (i=0; i < pldptr->numParmsToLoopOver; i++) {
    pldptr->parmCounter[i] = 0;
    PLsetParmValue(pldptr, i, pldptr->parmCounter[i]);
  }
}


/*
 * Move to the next set of parameter values
 */
void
PLincrementVars(PLdata *pldptr)
{
  int parmInd, doingParmChange;

  parmInd = 0;
  if (pldptr->numParmsToLoopOver > 0)
    doingParmChange = 1;
  else
    doingParmChange = 0;
  while (doingParmChange) {
    pldptr->parmCounter[parmInd]++;
    if (pldptr->parmCounter[parmInd] == pldptr->numVals[parmInd]) {
      /* reached the end of the set of values for the current parameter;
       * reset it back to the beginning and advance the next parameter */
      pldptr->parmCounter[parmInd] = 0;
      PLsetParmValue(pldptr, parmInd, pldptr->parmCounter[parmInd]);
      parmInd++;
      if (parmInd == pldptr->numParmsToLoopOver)
	doingParmChange = 0;
    }
    else {
      PLsetParmValue(pldptr, parmInd, pldptr->parmCounter[parmInd]);
      doingParmChange = 0;
    }
  }
}


/*
 * Print out a single parameter -- its name, type, and values.
 * All info is prefixed by the specified prefix string.
 */
void
PLprintSingleParm(PLdata *pldptr, int parmIndex, FILE *fp, char *prefix)
{
  char prefixStr[256];
  int i;

  if (pldptr->parmType[parmIndex] == DOUBLE) {
    if (pldptr->haveValueList[parmIndex]) {
      fprintf(fp, "%s   %s [double] has values", prefix, pldptr->parmName[parmIndex]);
      for (i=0; i < pldptr->numVals[parmIndex]; i++)
	fprintf(fp, " %g", pldptr->dValues[parmIndex][i]);
      fprintf(fp, "\n");
    } else {
      fprintf(fp, "%s   %s [double] from %g to %g with %d values (stepSize=%g)\n",
	      prefix, pldptr->parmName[parmIndex], pldptr->dbegin[parmIndex], pldptr->dend[parmIndex],
	      pldptr->numVals[parmIndex], pldptr->dstepSize[parmIndex]);
    }
  }
  else if (pldptr->parmType[parmIndex] == INT) {
    if (pldptr->haveValueList[parmIndex]) {
      fprintf(fp, "%s   %s [int] has values", prefix, pldptr->parmName[parmIndex]);
      for (i=0; i < pldptr->numVals[parmIndex]; i++)
	fprintf(fp, " %d", pldptr->iValues[parmIndex][i]);
      fprintf(fp, "\n");
    } else {
      fprintf(fp, "%s   %s [int] from %d to %d with %d values (stepSize=%d)\n",
	      prefix, pldptr->parmName[parmIndex], pldptr->ibegin[parmIndex], pldptr->iend[parmIndex],
	      pldptr->numVals[parmIndex], pldptr->istepSize[parmIndex]);
    }
  }
  if (pldptr->linkedParm[parmIndex] != NULL) {
    if (strstr(prefix, "--> linked") == NULL)
      snprintf(prefixStr, 256, "%s --> linked ", prefix);
    else
      strncpy(prefixStr, prefix, 256);
    PLprintSingleParm(pldptr->linkedParm[parmIndex], 0, fp, prefixStr);
  }
}


/*
 * Print out all of the varying parameters.  All info is prefixed with the specified commentChar
 */
void
PLprintParms(PLdata *pldptr, FILE *fp, char commentChar)
{
  int i;
  char prefixStr[5];

  sprintf(prefixStr, "%c", commentChar);
  if (pldptr->numParmsToLoopOver == 1)
    fprintf(fp, "%c There is 1 independently-varying parameter, ", commentChar);
  else if (pldptr->numParmsToLoopOver > 0)
    fprintf(fp, "%c There are %d independently-varying parameters, ", commentChar, pldptr->numParmsToLoopOver);
  if (pldptr->numParmsToLoopOver > 0) {
    fprintf(fp, "with a total of %d values.\n", pldptr->totalParmCombos);
    if (pldptr->numLinkedParms) {
      if (pldptr->numLinkedParms == 1)
	fprintf(fp, "%c There is also 1 linked parameter.\n", commentChar);
      else
	fprintf(fp, "%c There are also %d linked parameter.\n", commentChar, pldptr->numLinkedParms);
    }
    fprintf(fp, "%c Parameters are:\n", commentChar);
    for (i=0; i < pldptr->numParmsToLoopOver; i++) {
      PLprintSingleParm(pldptr, i, fp, prefixStr);
    }
  }
}
