/* Test the facility for looping over an arbitrary number of parameters,
 * on an OpenMPI cluster (i.e. multiple nodes going through the list of
 * parameter combos)
 * 
 * Example of usage:
 *  mpiexec -np 4 ./testparmloopMPI -phi 1,3.7,4.44 -mu 1:2:3@phi -alpha .32:.47:4 -bs 17,13,444,1234@alpha -reps 2
 *   That uses:
 *    - the three values 1, 3.7, and 4.44 for phi
 *      - the three values 1, 1.5, and 2 for mu, linked to the corresponding values of phi above
 *    - the four values 0.32, 0.37, 0.42, and 0.47 for alpha
 *      - the four values 17, 13, 444, and 1234 for blockSize, linked to the above values of alpha
 *    - the two values 1 and 2 for repNum
 *   This gives a total of 24 different parameter combos (3*4*2)
 *
 * David Hiebeler
 * http://www.math.umaine.edu/~hiebeler
 * May 2011
 *
Copyright (c) 2013 David Hiebeler. All Rights Reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1. Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

3. The name of the author may not be used to endorse or promote
products derived from this software without specific prior written
permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */

#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "getnumorrange.h"
#include "parmloop.h"


void
printusage(char *s)
{
  fprintf(stderr, "Usage: %s [-phi dz] [-mu dz] [-alpha dz] [-bs iz]\n", s);
  fprintf(stderr, "  dz and iz take one of the following forms (dz for doubles, iz for ints):\n");
  fprintf(stderr, "       value\n");
  fprintf(stderr, "       begin:end:numVals\n");
  fprintf(stderr, "       begin:end::stepSize\n");
  fprintf(stderr, "       val1,val2,val3,val4\n");
  fprintf(stderr, "       begin:end:numVals@otherParm   [to link with other specified parameter]\n");
  fprintf(stderr, "       begin:end::stepSize@otherParm   [to link with other specified parameter]\n");
  fprintf(stderr, "       val1,val2,val3,val4@otherParm  [to link with other specified parameter]\n");
  fprintf(stderr, "E.g.: %s -phi 1:8:10 -mu 0.2:3.4:10@phi -alpha 0.7\n", s);
  exit(2);
}


int
main(int argc, char **argv)
{
  double phi=0.0, mu=0.0, alpha=0.0;
  int blockSize=0, i, MPIrank, MPIsize, parmLoopInd, repNum=1;
  PLdata pldat;
  FILE *outFP;
  char outfname[256];

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &MPIrank);
  MPI_Comm_size(MPI_COMM_WORLD, &MPIsize);
  sprintf(outfname, "testparmloopMPI.out.node%d", MPIrank);
  if ((outFP=fopen(outfname, "w"))==NULL) {
    fprintf(stderr, "Unable to open file `%s' for writing", outfname);
    return(1);
  }
  i = 1;
  PLinitData(&pldat);
  while (i < argc) {
    if (!strcmp(argv[i], "-phi")) {
      if (i >= argc-1)
	printusage(argv[0]);
      if (PLreadDoubleArg(argv[i+1], &pldat, (void *)&phi, "phi"))
	printusage(argv[0]);
      i += 2;
    }
    else if (!strcmp(argv[i], "-mu")) {
      if (i >= argc-1)
	printusage(argv[0]);
      if (PLreadDoubleArg(argv[i+1], &pldat, (void *)&mu, "mu"))
	printusage(argv[0]);
      i += 2;
    }
    else if (!strcmp(argv[i], "-alpha")) {
      if (i >= argc-1)
	printusage(argv[0]);
      if (PLreadDoubleArg(argv[i+1], &pldat, (void *)&alpha, "alpha"))
	printusage(argv[0]);
      i += 2;
    }
    else if (!strcmp(argv[i], "-bs")) {
      if (i >= argc-1)
	printusage(argv[0]);
      if (PLreadIntArg(argv[i+1], &pldat, (void *)&blockSize, "blockSize"))
	printusage(argv[0]);
      i += 2;
    }
    else if (!strcmp(argv[i], "-reps")) {
      if (i >= argc-1)
	printusage(argv[0]);
      if (PLreadIntArgValueAsCounter(argv[i+1], &pldat, (void *)&repNum, "repetitions"))
	printusage(argv[0]);
      i += 2;
    }
    else
      printusage(argv[0]);
  }
  PLinitVars(&pldat);
  if (MPIrank == 0)
    PLprintParms(&pldat, stdout, '%');

  for (parmLoopInd=0; parmLoopInd < pldat.totalParmCombos; parmLoopInd++) {
    if ((parmLoopInd%MPIsize) == MPIrank)
      fprintf(outFP, "[node %d] %d: phi=%g, mu=%g, alpha=%g, blocksize=%d, repNum=%d\n", MPIrank,
	     parmLoopInd, phi, mu, alpha, blockSize, repNum);

    /*
     *
     * Put actual simulation code here
     *
     */

    PLincrementVars(&pldat);
  }
  fclose(outFP);
  MPI_Finalize();
  return(0);
}
